SECONDS_IN_ONE_DAY = 60*60*24 # 86400  # used for granularity (daily)


import logging
logger = logging.getLogger('isitfit')


# Exception classes
class NoCloudtrailException(Exception):
    pass

class DdgNoData(ValueError):
  pass

class HostNotFoundInDdg(DdgNoData):
  pass

class DataNotFoundForHostInDdg(DdgNoData):
  pass

class NoCloudwatchException(Exception):
    pass

class IsitfitCliRunnerBreakIterator(Exception):
  pass


def mergeSeriesOnTimestampRange(df_cpu, df_type, fields):
  """
  Upsamples df_type to df_cpu.
  Check unit test for an example
  """
  # check that df_type dataframe are sorted descending
  # This does not apply to df_cpu
  #if df_cpu.Timestamp.iloc[0] < df_cpu.Timestamp.iloc[-1]:
  #  raise Exception("CPU Dataframe should be sorted descending for utils.mergeSeriesOnTimestampRange")

  if df_type.iloc[0].name < df_type.iloc[-1].name:
    raise Exception("Types Dataframe should be sorted descending for utils.mergeSeriesOnTimestampRange")

  import numpy as np

  for f in fields:
    df_cpu[f] = None

  # assume df_type is sorted in decreasing EventTime order (very important)
  # NB: since some instances are not present in the cloudtrail (for which we append artificially the "now" type)
  #     Need to traverse the df_type matrix backwards
  for index, row_type in df_type.iterrows():
    row_i = np.where(df_cpu.Timestamp <= row_type.name)[0]
    for f in fields:
      # use row_type.name instead of row_type['EventTime']
      # check note above about needing to traverse backwards
      # df_cpu.iloc[np.where(df_cpu.Timestamp >= row_type.name)[0], df_cpu.columns.get_loc('instanceType')] = row_type['instanceType']
      col_i = df_cpu.columns.get_loc(f)
      df_cpu.iloc[row_i, col_i] = row_type[f]

  # fill na at beginning with back-fill
  # (artifact of cloudwatch having data at days before the creation of the instance)
  for f in fields:
    df_cpu[f] = df_cpu[f].fillna(method='backfill')

  return df_cpu







# copied from git-remote-aws
def mysetlocale():
  li = 'en_US.utf8'
  import os
  os.environ["LC_ALL"] = li
  os.environ["LANG"]   = li




MAX_ROWS = 10
MAX_COLS = 5
MAX_STRING = 20
def display_df(title, df, csv_fn, shape, logger):
    # https://pypi.org/project/termcolor/
    from termcolor import colored
    import click

    click.echo("")

    if shape[0]==0:
      click.echo(title)
      click.echo(colored("None", "red"))
      return

    if csv_fn is not None:
      click.echo(colored("The table '%s' was saved to the CSV file '%s'."%(title, csv_fn), "cyan"))
      click.echo(colored("It could be opened in the terminal with visidata (http://visidata.org/)","cyan"))
      click.echo(colored("and you can close visidata by pressing 'q'","cyan"))
      open_vd = input(colored('Would you like to do so? yes/[no] ', 'cyan'))
      if open_vd.lower() == 'yes' or open_vd.lower() == 'y':
        click.echo("Opening CSV file `%s` with visidata."%csv_fn)
        from subprocess import call
        call(["vd", csv_fn])
        click.echo("Exited visidata.")
        click.echo(colored("The table '%s' was saved to the CSV file '%s'."%(title, csv_fn), "cyan"))
        return
      else:
        click.echo("Not opening visidata.")
        click.echo("To open the results with visidata, use `vd %s`."%csv_fn)


    # if not requested to open with visidata
    from tabulate import tabulate
    df_show = df.head(n=MAX_ROWS)
    df_show = df_show.applymap(lambda c: (c[:MAX_STRING]+'...' if len(c)>=MAX_STRING else c) if type(c)==str else c)

    click.echo(tabulate(df_show, headers='keys', tablefmt='psql', showindex=False))

    if (shape[0] > MAX_ROWS) or (shape[1] > MAX_COLS):
      click.echo("...")
      click.echo("(results truncated)")
      # done
      return

    # done
    return


def prompt_upgrade(pkg_name, current_version):
  """
  check if current version is out-of-date
  https://github.com/alexmojaki/outdated

  copied from https://github.com/WhatsApp/WADebug/blob/958ac37be804cc732ae514d4872b93d19d197a5c/wadebug/cli.py#L40
  """
  import outdated

  is_outdated = False
  try:
    is_outdated, latest_version = outdated.check_outdated(pkg_name, current_version)
  except requests.exceptions.ConnectionError as error:
    # e.g.
    # requests.exceptions.ConnectionError:
    # HTTPSConnectionPool(host='pypi.python.org', port=443):
    # Max retries exceeded with url: /pypi/isitfit/json
    # (Caused by NewConnectionError(... Name or service not known)
    # Thrown upon internet disconnection
    pass
  except ValueError as error:
    # catch case of "ValueError: Version 0.10.0 is greater than the latest version on PyPI: 0.9.1"
    # This would happen on my dev machine
    if not "is greater than" in str(error):
      raise

    # In this case, outdated does not cache the result to disk
    # so cache it myself (copied from https://github.com/alexmojaki/outdated/blob/565bb3fe1adc30da5e50249912cd2ac494662659/outdated/__init__.py#L61)
    latest_version = str(error).split(":")[1].strip()
    import datetime as dt
    import json
    with outdated.utils.cache_file(pkg_name, 'w') as f:
      try:
        data = [latest_version, outdated.utils.format_date(dt.datetime.now())]
        json.dump(data, f)
      except Exception as e:
        print('Error: ' + str(e))
        raise


  # is_outdated = True # FIXME for debugging
  if not is_outdated:
      return is_outdated

  import click
  msg_outdated = """The current version of {pkg_name} ({current_version}) is out of date.
Run `pip3 install {pkg_name} --upgrade` to upgrade to version {latest_version},
or use `isitfit --skip-check-upgrade ...` to skip checking for version upgrades of isitfit.
"""
  msg_outdated = msg_outdated.format(
      pkg_name=pkg_name, current_version=current_version, latest_version=latest_version
    )
  click.secho(msg_outdated, fg="red")

  # Give the user some time to read the message and possibly update
  import time
  from tqdm import tqdm
  wait_outdated = 10
  click.secho("Will continue in %i seconds"%wait_outdated, fg='yellow')
  for i in tqdm(range(wait_outdated)):
    time.sleep(1)

  return is_outdated


# This import needs to stay here for the sake of the mock in test_utils
import requests
SKIP_PING=False
def ping_matomo(action_name):
  """
  Gather anonymous usage statistics
  """
  logger.debug("ping_matomo('%s')"%action_name)

  # get uuid
  from .dotMan import DotMan
  uuid_val = DotMan().get_myuid()

  # get version
  from . import isitfit_version as isitfit_cli_version

  # build action name field. note that "action_name" already starts with "/"
  full_actionName = "%s%s"%(isitfit_cli_version, action_name)

  # use base function
  from matomo_sdk_py.matomo_sdk_py import ping_matomo as ping_matomo_base
  ping_matomo_base(
    action_name=full_actionName,
    action_base="https://cli.isitfit.io",
    idsite=2, # 2 is for cli.isitfit.io
    uuid_val=uuid_val,
    matomo_url="https://isitfit.matomo.cloud/piwik.php"
  )


def display_footer():
    import click

    from . import isitfit_version

    click.echo("")
    click.echo("⛅ Generated by isitfit version %s"%isitfit_version)
    click.echo("")
    click.echo("Useful links:")
    click.echo("ℹ️  isitfit homepage         https://isitfit.autofitcloud.com")
    click.echo("😞 isitfit issues           https://github.com/autofitcloud/isitfit/issues")
    click.echo("🌎 Global Climate Strike    https://twitter.com/hashtag/ClimateStrike")
    click.echo("❤️  Built by AutofitCloud    https://www.autofitcloud.com")





def myreturn(df_xxx):
    if df_xxx.shape[0] > 0:
      return df_xxx
    else:
      return None # this means that the data was found in cache, but it was empty (meaning aws returned no data)



def b2l(b_in):
  """
  return "T" on true and "F" on false
  Instead of "True" and "False"
  """
  return str(b_in)[0]


def l2s(x):
  """
  [1,2,3,4,5,6,7,8] -> '1,2,...,7,8'
  """
  if len(x)>5: x = x[:2] + ['...'] + x[-2:]
  y = [str(z) for z in x] # convert to list of strings, eg if list of int
  y = ",".join(y)
  return y



def taglist2str(taglist, filter_tags):
  """
  eg
  taglist2str([{'Key':'app', 'Value':'isitfit'}], 'boo')
  returns ""

  taglist2str([{'Key':'app', 'Value':'isitfit'}], 'is')
  returns "app = isitfit"
  """
  if filter_tags is not None:
    # filter the tag list for only those containing the filter-tags string
    f_tn = filter_tags.lower()

    # similar to the isitfit.mainManager.tagsContain function, but filtering the tags themselves
    taglist = [x for x in taglist if (f_tn in x['Key'].lower()) or (f_tn in x['Value'].lower())]

  # list to string
  taglist = ["%s = %s"%(x['Key'], x['Value']) for x in taglist]
  taglist = "\n".join(taglist)

  return taglist


def pd_series_frozenset_union(s1, s2):
  """
  Pandas doesn't have a built-in set union.
  Worse, it doesn't support set as column type
  Need to use frozenset
  Check test for example
  Ref https://python-forum.io/Thread-Error-Message-TypeError-unhashable-type-set
  """
  import pandas as pd
  df1=pd.DataFrame({'a1': s1})
  df2=pd.DataFrame({'a2': s2})
  df3=df1.merge(df2, left_index=True, right_index=True, how='left')
  df3['a3'] = df3.apply(lambda row: row.a1 if pd.isnull(row.a2) else row.a1.union(row.a2), axis=1)
  return df3['a3']



class AwsProfileMan:
  def __init__(self):
    import boto3
    self.profile_list_nocolors = boto3.session.Session().available_profiles

    from isitfit.dotMan import DotLastProfile
    self.last_profile_cls = DotLastProfile()

    self.w2c = Word2Color()


  def validate_profile(self, ctx, param, value_colored):
    if value_colored is None: return value_colored

    # strip color from value_colored
    # http://stackoverflow.com/questions/14693701/ddg#14693789
    import re
    ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')
    value_nocolor = ansi_escape.sub('', value_colored)

    # check if in list
    if value_nocolor not in self.profile_list_nocolors:
      import click
      err_m = 'Profile %s is not from ~/.aws/credentials file.'%value_colored
      raise click.BadParameter(err_m)

    # set the profile in an env var so that boto3 picks it up automatically
    if value_nocolor is not None:
      import os
      os.environ['AWS_PROFILE'] = value_nocolor

    # save profile in click context for other usage in displayed/emailed report
    ctx.obj['aws_profile'] = value_nocolor

    # save in last-used profile file
    self.last_profile_cls.set(value_nocolor)

    # done
    return value_nocolor


  def prompt(self):
    x = []
    x.append("Profiles in AWS credential file:")

    from termcolor import colored
    profile_list_colors = [colored(z, self.w2c.convert(z)) for z in self.profile_list_nocolors]

    x += ["- %s"%z for z in profile_list_colors]
    x.append("")
    x.append("(use `AWS_PROFILE=myprofile isitfit ...` or `isitfit command --profile=myprofile ...` to skip this prompt)")
    x.append("Profile to use")
    y = "\n".join(x)
    return y


  def default(self):
    from termcolor import colored
    z = self.last_profile_cls.get() or 'default'
    val_color = colored(z, self.w2c.convert(z))
    return val_color



class PromptToEmailIfNotRequested():
  def __init__(self):
    from isitfit.dotMan import DotLastEmail
    self.last_email_cl = DotLastEmail()
    self.last_email_val = None

    # more quick validation
    # works with a@b.c but not a@b@c.d
    # https://stackoverflow.com/questions/8022530/how-to-check-for-valid-email-address#8022584
    import re
    self.EMAIL_REGEX = re.compile(r"[^@]+@[^@]+\.[^@]+")

  def _helper(self):
      # use "default=''" so that the "leave blank to skip" works (instead of click re-prompting until it gets a value)
      import click
      res_prompt = click.prompt('Email to which to share the results (type "skip" to not send email)', type=str, default=self.last_email_val)

      # check if blank
      res_prompt = res_prompt.strip()
      if res_prompt=='skip':
        ping_matomo("/cost/share_email?original=F&provided=F")
        return None

      # quick validate
      # shortest email is: a@b.c
      # Longest email is: shadishadishadishadi@shadishadishadishadi.shadi
      if len(res_prompt) >= 5:
        if len(res_prompt) <= 50:
          if bool(self.EMAIL_REGEX.match(res_prompt)):
            ping_matomo("/cost/share_email?original=F&provided=T")
            return [res_prompt]

      # otherwise, invalid email
      logger.error("Invalid email address: '%s'"%res_prompt)
      raise ValueError


  def prompt(self, emailTo):
    from isitfit.utils import ping_matomo

    if emailTo is not None:
      if len(emailTo) > 0:
        # user already requested email
        ping_matomo("/cost/share_email?original=T")
        return emailTo

    # get last used email if available
    self.last_email_val = self.last_email_cl.get()
    if self.last_email_val is None: self.last_email_val='skip'

    # prompt for email
    while True:
      try:
        res = self._helper()
        return res
      except ValueError:
        pass



class Word2Color:
    """
    Calculate a color (from termcolor package) for a word.
    Generates the same color for the same word.
    https://github.com/hfeeki/termcolor/blob/master/termcolor.py
    """
    def __init__(self):
      from termcolor import COLORS
      self.colors = sorted(list(COLORS.keys()))
      self.colors = [x for x in self.colors if x!='white']
      self.n_colors = len(self.colors)

    def convert(self, word):
      hash_ = sum([ord(letter) for letter in word])
      hash_ = hash_ % self.n_colors
      #print("word", word, "hash", hash_)
      return self.colors[hash_]



def pd_subset_latest(df1, field_val, field_sortmax):
    """
    # only keep the latest section with the latest size
    # (eg if the ec2 had the size s1 then s2 then s1, only keep the latest)
    """
    if df1.shape[0]==0: return df1

    df2 = df1.sort_values(by=field_sortmax, ascending=True)
    latest_val = df2[field_val].iloc[-1]

    latest_idx = (df2[field_val] == latest_val).astype(int)
    latest_idx = latest_idx.sort_index(ascending=False).cummin().sort_index(ascending=True)

    df3 = df2[latest_idx.astype(bool)]
    return df3



"""
Classes to help with managing visibility of tqdm progress bars at different levels of the code
https://github.com/tqdm/tqdm/
"""

class TqdmL2Base:
  def __init__(self):
    # override to display tqdm
    self.show = False

  def __call__(self, iterator, *args, **kwargs):
    if not self.show:
      return iterator

    from tqdm import tqdm
    return tqdm(iterator, *args, **kwargs)


class TqdmL2Quiet(TqdmL2Base):
  """
  Sets show=True when isitfit used without --verbose nor --debug, i.e. isitfit cost analyze
  Useful for showing a service-level progressbar, i.e. moves by 1 for ec2, then 1 for redshift, etc
  The problem is that currently only ec2 and redshift are supported,
  which makes this progressbar lose its meaning.
  """
  def __init__(self, ctx):
    # FIXME update 2019-12-09
    #   Until I build a meaningful progressbar at the level of services,
    #   this progressbar is getting completely disabled.
    # self.show = not(ctx.obj['debug'] or ctx.obj['verbose'])
    self.show = False


class TqdmL2Verbose(TqdmL2Base):
  """
  Sets show=True when isitfit used with either --verbose or --debug, i.e. isitfit --verbose cost analyze
  """
  def __init__(self, ctx):
    # FIXME update 2019-12-09
    # Related to the note in TqdmL2Quiet, this progressbar is getting enabled now all the time, until a proper service-level progress bar is implemented
    # self.show = ctx.obj['debug'] or ctx.obj['verbose']
    self.show = True




# RuntimeError: Click will abort further execution because Python 3 was configured to use ASCII as encoding for the environment.
# Consult https://click.palletsprojects.com/en/7.x/python3/ for mitigation steps.
#
# Edit 2019-10-08: whatsapp's wadebug uses "click.disable_unicode_literals_warning = True"
# Ref: https://github.com/WhatsApp/WADebug/blob/958ac37be804cc732ae514d4872b93d19d197a5c/wadebug/cli.py#L23
from ..utils import mysetlocale
mysetlocale()


from isitfit.utils import logger


import click

from .. import isitfit_version


# For the --share-email "multiple options"
# https://click.palletsprojects.com/en/7.x/options/#multiple-options
from isitfit.cli.click_descendents import isitfit_group
@isitfit_group(invoke_without_command=False)
@click.option('--debug', is_flag=True, help='Display more details to help with debugging')
@click.option('--verbose', is_flag=True, help='Display more details to help with understanding program flow')
@click.option('--optimize', is_flag=True, help='DEPRECATED: use "isitfit cost optimize" instead', hidden=True)
@click.option('--version', is_flag=True, help='DEPRECATED: use "isitfit version" instead', hidden=True)
@click.option('--share-email', multiple=True, help='Share result to email address')
@click.option('--skip-check-upgrade', is_flag=True, help='Skip step for checking for upgrade of isitfit')
@click.option('--skip-prompt-email', is_flag=True, help='Skip the prompt to share result to email address')
@click.pass_context
def cli_core(ctx, debug, verbose, optimize, version, share_email, skip_check_upgrade, skip_prompt_email):
    # FIXME click bug: `isitfit cost --help` is calling the code in here. Workaround is to check --help
    import sys
    if '--help' in sys.argv: return

    # make sure that context is a dict
    ctx.ensure_object(dict)

    # set up exception aggregation in sentry.io
    from isitfit import sentry_proxy
    from isitfit.apiMan import BASE_URL
    sp_url = f"{BASE_URL}fwd/sentry"
    sentry_proxy.init(dsn=sp_url)

    # test exception caught by sentry. FIXME Dont commit this! :D
    # 1/0

    # usage stats
    # https://docs.python.org/3.5/library/string.html#format-string-syntax
    from isitfit.utils import ping_matomo, b2l
    ping_url = "/?debug={}&verbose={}&share_email={}&skip_check_upgrade={}"
    ping_url = ping_url.format(b2l(debug), b2l(verbose), b2l(len(share_email)>0), b2l(skip_check_upgrade))
    ping_matomo(ping_url)

    # choose log level based on debug and verbose flags
    import logging
    logLevel = logging.DEBUG if debug else (logging.INFO if verbose else logging.WARNING)

    ch = logging.StreamHandler()
    ch.setLevel(logLevel)
    logger.addHandler(ch)
    logger.setLevel(logLevel)

    if debug:
      logger.debug("Enabled debug level")
      logger.debug("-------------------")

    # After adding the separate command for "cost" (i.e. `isitfit cost analyze`)
    # putting a note here to notify user of new usage
    # Ideally, this code would be deprecated though
    if ctx.invoked_subcommand is None:
        # if still used without subcommands, notify user of new usage
        #from .cost import analyze as cost_analyze, optimize as cost_optimize
        #if optimize:
        #  ctx.invoke(cost_optimize, filter_tags=filter_tags, n=n)
        #else:
        #  ctx.invoke(cost_analyze, filter_tags=filter_tags)
        from click.exceptions import UsageError
        if optimize:
          err_msg = "As of version 0.11, please use `isitfit cost optimize` instead of `isitfit --optimize`."
          ping_matomo("/error/UsageError?message=%s"%err_msg)
          raise UsageError(err_msg)
        elif version:
          # ctx.invoke(cli_version)
          err_msg = "As of version 0.11, please use `isitfit version` instead of `isitfit --version`."
          ping_matomo("/error/UsageError?message=%s"%err_msg)
          raise UsageError(err_msg)
        else:
          err_msg = "As of version 0.11, please use `isitfit cost analyze` instead of `isitfit` to calculate the cost-weighted utilization."
          ping_matomo("/error/UsageError?message=%s"%err_msg)
          raise UsageError(err_msg)

    # check if emailing requested
    if share_email is not None:
      max_n_recipients = 3
      if len(share_email) > max_n_recipients:
          err_msg = "Maximum allowed number of email recipients is %i. Received %i"%(max_n_recipients, len(share_email))
          ping_matomo("/error?message=%s"%err_msg)
          from click.exceptions import BadParameter
          raise BadParameter(err_msg, param_hint="--share-email")

      ctx.obj['share_email'] = share_email

    # check if current version is out-of-date
    if ctx.invoked_subcommand != 'version':
      if not skip_check_upgrade:
        from ..utils import prompt_upgrade
        is_outdated = prompt_upgrade('isitfit', isitfit_version)
        ctx.obj['is_outdated'] = is_outdated
        if is_outdated:
          ping_matomo("/version/prompt_upgrade?is_outdated=%s"%b2l(is_outdated))


    if ctx.invoked_subcommand not in ['version', 'migrations']:
      # run silent migrations
      from isitfit.migrations.migman import silent_migrate
      migname_l = silent_migrate()
      if len(migname_l)>0:
        from isitfit.utils import l2s
        migname_s = l2s(migname_l)
        ping_matomo("/migrations/silent?migname=%s"%(migname_s))


    # save `verbose` and `debug` for later tqdm
    ctx.obj['debug'] = debug
    ctx.obj['verbose'] = verbose
    # save skip-prompt-email for later usage
    ctx.obj['skip_prompt_email'] = skip_prompt_email



from .tags import tags as cli_tags
from .cost import cost as cli_cost
from .version import version as cli_version
from isitfit.migrations.cli import migrations as cli_migrations
from .issue10 import issue10 as cli_issue10

cli_core.add_command(cli_version)
cli_core.add_command(cli_cost)
cli_core.add_command(cli_tags)
cli_core.add_command(cli_migrations)
cli_core.add_command(cli_issue10)


#-----------------------

if __name__ == '__main__':
  cli_core()



"""
Configure the sentry_sdk package to send HTTP requests to a custom endpoint
instead of the normal https://foo@sentry.io/1234 DSN

This is useful to intercept the sentry HTTP requests on a server
and then forward to the original sentry.io endpoint.
This way, the sentry key need not be exposed to the original client,
as would be the case if I use sentry_sdk directly in isitfit-cli.

Similar: https://github.com/getsentry/sentry/issues/1230

Usage:
  import sentry_proxy
  sentry_proxy.init(dsn="http://webhook.site/1234")
  
Tests:
  pytest isitfit/test_sentryProxy.py -k test_sentry_proxy_fake
  pytest isitfit/test_sentryProxy.py -k test_sentry_proxy_isitfit
"""

import sentry_sdk

# Setting the DSN doesn't work because of the validation
# SENTRY_SDK = "http://foo@webhook.site/5716352c-308a-45f1-b862-92ab80560cec"
# sentry_sdk.init(SENTRY_SDK)
#----------------------
# Create a derived sentry Dsn object
# since sentry_sdk.utils.Dsn.__init__ uses isinstance
# >>> class A: pass
# >>> class B(A): pass
# >>> x=B()
# >>> isinstance(x,B)
# True
# >>> isinstance(x,A)
# True
#-------------
from sentry_sdk.utils import Dsn, Auth
from sentry_sdk.transport import HttpTransport
import urllib.parse as urlparse


class MyAuth(Auth):
    @property
    def store_api_url(self):
      return self.real_dsn


class MyDsn(Dsn):
  def __init__(self, real_dsn):
    fake_dsn = "https://foo@sentry.io/1234"
    super().__init__(fake_dsn)

    parts = urlparse.urlsplit(real_dsn)
    self.scheme = parts.scheme
    self.host = parts.hostname
    self.port = parts.port or (parts.scheme == "https" and 443 or 80)
    self.public_key = None
    self.secret_key = None
    self.project_id = None
    self.path = parts.path
    self.real_dsn = real_dsn # new variable for the sake of to_auth below


  def to_auth(self, client=None):
        # over-write function
        auth = super().to_auth(client)
        # add member
        auth.real_dsn = self.real_dsn
        # cast class to use new member
        auth.__class__ = MyAuth
        return auth


class MyTransport(HttpTransport):
  def __init__(self, options=None):
    super().__init__(options)
    # cast http://stackoverflow.com/questions/3464061/ddg#3464154
    self.parsed_dsn.__class__ = MyDsn
    # save self._auth again
    self._auth = self.parsed_dsn.to_auth("sentry-proxy.python/v0")


def init(dsn):
    # use default_integrations=False to avoid the AtexitIntegration from displaying a message
    # "Senty is attempting to send ..."
    # https://docs.sentry.io/platforms/python/default-integrations/
    parsed_dsn = MyDsn(dsn)
    sentry_sdk.init(dsn=parsed_dsn, transport=MyTransport, default_integrations=False)




from isitfit.utils import logger



class ServiceIterator:
  """
  Similar to isitfit.cost.redshift.iterator.BaseIterator
  """
  service_description = "AWS EC2, Redshift"
  region_include = None
  def __init__(self, s_ec2, s_redshift):
    self.s_ec2 = s_ec2
    self.s_redshift = s_redshift

  def count(self):
    return 2

  def get_regionInclude(self):
    return None

  def __iter__(self):
    s_l = [ ('ec2',      self.s_ec2),
            ('redshift', self.s_redshift)
          ]
    for s_name, s_i in s_l:
      yield None, s_name, None, s_i



class ServiceCalculatorGet:
  def per_service(self, context_service):
    service_i = context_service['ec2_obj']

    # configure tqdm
    from isitfit.tqdmman import TqdmL2Verbose
    tqdml2_obj = TqdmL2Verbose(service_i.ctx)

    # run pipeline
    context_all = service_i.get_ifi(tqdml2_obj)
    context_service['context_all'] = context_all
    return context_service


class ServiceCalculatorSave:
  def __init__(self):
    self.table_d = {}

  def per_service(self, context_service):
    service_name = context_service['ec2_id']

    context_all = context_service['context_all']
    if context_all is None: return context_service

    table_single = context_all['table']
    table_single = {v['label']: v for v in table_single}
    self.table_d[service_name] = table_single

    return context_service

    

class ServiceCalculatorBinned:
  def __init__(self):
    self.dfbin_d = {}
    self.dfbin_p = None

  def per_service(self, context_service):
    service_name = context_service['ec2_id']

    context_all = context_service['context_all']
    if context_all is None: return context_service

    # get df_bin
    if 'df_bins' not in context_service['context_all']:
      return context_service

    # shortcut
    df_i = context_service['context_all']['df_bins']

    # timestamp index to regular column to convert to date
    # Update 2019-12-08 need to convert to string instead of date
    # because otherwise in --share-email
    # the shared email will show the UNIX integer timestamp in the column name
    # and the date cells
    df_i = df_i.reset_index()
    for fx in ['Timestamp', 'dt_start', 'dt_end']:
      # df_i[fx] = df_i[fx].dt.date
      df_i[fx] = df_i[fx].dt.strftime("%Y-%m-%d")

    df_i.set_index('Timestamp', inplace=True)

    # sort columns for later's after_all, and drop regions_set
    df_i = df_i[['dt_start', 'dt_end', 'regions_str', 'count_analyzed', 'capacity_usd', 'used_usd', 'used_pct']]
    
    # pretty-print dollar/percentage signs
    df_i['capacity_usd'] = df_i['capacity_usd'].apply(lambda x: "%0.0f $"%x)
    df_i['used_usd'    ] = df_i['used_usd'    ].apply(lambda x: "%0.0f $"%x)
    df_i['used_pct'    ] = df_i['used_pct'    ].apply(lambda x: "%0.0f %%"%x)

    
    # Replace code-friendly strings with human-friendly strings
    fm = {
      'dt_start': 'Start date',
      'dt_end': 'End date',
      'regions_str': 'Regions',
      'count_analyzed': 'Resources analyzed',
      'capacity_usd': 'Billed cost',
      'used_usd': 'Used cost',
      'used_pct': 'CWAU (Used/Billed)',
    }
    df_i.rename(columns=fm, inplace=True)

    # transpose to match format of one-column output
    df_i = df_i.transpose()

    # add service name, so that EC2 and Redshift don't get mixed up
    df_i['Service'] = service_name

    # save
    self.dfbin_d[service_name] = df_i

    return context_service


  def after_all(self, context_all):
    """
    concatenate the self.dfbin_d dict to pandas dataframe
    """
    dfbin_l = [v.reset_index() for k,v in self.dfbin_d.items()]
    if len(dfbin_l)==0:
      from isitfit.cli.click_descendents import IsitfitCliError
      raise IsitfitCliError("No data found")

    import pandas as pd
    self.dfbin_p = pd.concat(dfbin_l, axis=0)

    # rename the weird "Timestamp" column to "Field"
    self.dfbin_p.rename(columns={'index': 'Field'}, inplace=True)

    # set index
    self.dfbin_p.set_index(['Service', 'Field'], inplace=True)

    # save for later processing
    context_all['dfbin_p'] = self.dfbin_p

    # done
    return context_all



from isitfit.cost.base_reporter import ReporterBase

#class ServiceReporterTotals(ReporterBase):
#  def __init__(self):
#    self.table_merged = []
#
#  def postprocess(self, context_all):
#    # get first available start/end date
#    analyzer = context_all['analyzer']
#    date_source = analyzer.table_d.get('ec2', analyzer.table_d.get('redshift', None))
#    if date_source is None:
#      # no data for ec2 and redshift
#      return context_all
#
#    self.table_merged += [
#      {'color': '',
#       'label': "Start date",
#       'value':  date_source['Start date']['value'] # just take first
#      },
#      {'color': '',
#       'label': "End date",
#       'value': date_source['End date']['value'] # just take first
#      },
#    ]
#
#    # if no EC2:
#    if 'ec2' not in analyzer.table_d.keys():
#      self.table_merged.append(
#        {'color': '',
#         'label': "EC2 instances",
#         'value': "None found",
#        },
#      )
#    else:
#      self.table_merged += [
#        {'color': '',
#         'label': "EC2 Regions",
#         'value': analyzer.table_d['ec2']['Regions']['value'],
#        },
#        {'color': '',
#         'label': "EC2 machines (total)",
#         'value': analyzer.table_d['ec2']['EC2 machines (total)']['value'] # only 1 anyway
#        },
#        {'color': '',
#         'label': "EC2 machines (analyzed)",
#         'value': analyzer.table_d['ec2']['EC2 machines (analyzed)']['value'] # only 1 anyway
#        },
#        {'color': 'cyan',
#         'label': "EC2 Billed cost",
#         'value': analyzer.table_d['ec2']['Billed cost']['value']
#        },
#        {'color': 'cyan',
#         'label': "EC2 Used cost",
#         'value': analyzer.table_d['ec2']['Used cost']['value']
#        },
#        {'color': analyzer.table_d['ec2']['CWAU (Used/Billed)']['color'],
#         'label': "EC2 CWAU (Used/Billed)",
#         'value': analyzer.table_d['ec2']['CWAU (Used/Billed)']['value']
#        },
#      ]
#
#    # if no Redshift:
#    if 'redshift' not in analyzer.table_d.keys():
#      self.table_merged.append(
#        {'color': '',
#         'label': "Redshift clusters",
#         'value': "None found",
#        },
#      )
#    else:
#      self.table_merged += [
#        {'color': '',
#         'label': "Redshift Regions",
#         'value': analyzer.table_d['redshift']['Regions']['value']
#        },
#        {'color': '',
#         'label': "Redshift clusters (total)",
#         'value': analyzer.table_d['redshift']['Redshift clusters (total)']['value'] # only 1 anyway
#        },
#        {'color': '',
#         'label': "Redshift clusters (analyzed)",
#         'value': analyzer.table_d['redshift']['Redshift clusters (analyzed)']['value'] # only 1 anyway
#        },
#        {'color': 'cyan',
#         'label': "Redshift Billed cost",
#         'value': analyzer.table_d['redshift']['Billed cost']['value']
#        },
#        {'color': 'cyan',
#         'label': "Redshift Used cost",
#         'value': analyzer.table_d['redshift']['Used cost']['value']
#        },
#        {'color': analyzer.table_d['redshift']['CWAU (Used/Billed)']['color'],
#         'label': "Redshift CWAU (Used/Billed)",
#         'value': analyzer.table_d['redshift']['CWAU (Used/Billed)']['value']
#        },
#    ]
#
#    # done
#    return context_all
#
#  def display(self, context_all):
#    import click
#
#    if not self.table_merged:
#      click.echo("No resources found in AWS EC2, Redshift")
#      return context_all
#
#    # https://pypi.org/project/termcolor/
#    from termcolor import colored
#
#    def get_row(row):
#        def get_cell(i):
#          retc = row[i] if not row['color'] else colored(row[i], row['color'])
#          return retc
#        
#        retr = [get_cell('label'), get_cell('value')]
#        return retr
#
#    dis_tab = [get_row(row) for row in self.table_merged]
#
#    # logger.info("Summary:")
#    from tabulate import tabulate
#    click.echo("Cost-Weighted Average Utilization (CWAU) of the AWS account:")
#    click.echo("")
#    click.echo(tabulate(dis_tab, headers=['Field', 'Value']))
#    click.echo("")
#    click.echo("For reference:")
#    click.echo(colored("* CWAU >= 70% is well optimized", 'green'))
#    click.echo(colored("* CWAU <= 30% is underused", 'red'))
#
#    return context_all
#
#  def email(self, context_all):
#      if self.emailTo is None: return context_all
#
#      context_2 = {}
#      context_2['emailTo'] = self.emailTo
#      context_2['click_ctx'] = context_all['click_ctx']
#      context_2['dataType'] = 'cost analyze' # redshift + ec2
#      context_2['dataVal'] = {'table': self.table_merged}
#      super().email(context_2)
#
#      return context_all



class ServiceReporterBinned(ReporterBase):
  def display(self, context_all):
    dfbin_p = context_all['dfbin_p']

    import click
    from termcolor import colored
    from tabulate import tabulate

    # logger.info("Summary:")
    click.echo("Cost-Weighted Average Utilization (CWAU) of the AWS account (EC2, Redshift):")
    click.echo("")
    click.echo(dfbin_p)
    click.echo("")
    click.echo("For reference:")
    click.echo(colored("* CWAU >= 70% is well optimized", 'green'))
    click.echo(colored("* CWAU <= 30% is underused", 'red'))
    click.echo("")

    return context_all


  def email(self, context_all):
      if self.emailTo is None: return context_all

      dfbin_p = context_all['dfbin_p']
      ctx = context_all['click_ctx']

      # FIXME pandas bug: to_json on dataframe with index yields unreadable json
      # i.e. pd.read_json(dfbin_p.to_json()) raises an exception
      # So need to reset_index first, and then set index again on the server-side
      # dfbin_s = dfbin_p.to_json()
      dfbin_s = dfbin_p.reset_index().to_json()

      context_2 = {}
      context_2['emailTo'] = self.emailTo
      context_2['click_ctx'] = context_all['click_ctx']
      context_2['dataType'] = 'cost analyze (binned)' # redshift + ec2
      context_2['dataVal'] = {
        'table': dfbin_s,
        'options': {
          'aws_profile': ctx.obj['aws_profile'],
          'filter_region': ctx.obj['filter_region'] or 'all'
        }
      }
      super().email(context_2)

      return context_all




def pipeline_factory(mm_eca, mm_rca, ctx, share_email):
    """
    Combines the 2 pipelines from EC2 and Redshift
    mm_eca - pipeline of EC2 cost analyze
    mm_rca - pipeline of Redshift cost analyze
    ctx - click context
    share_email - list of emails or None
    """
    from isitfit.cost.mainManager import RunnerAccount
    mm_all = RunnerAccount("AWS cost analyze (EC2, Redshift) in all regions", ctx)

    service_iterator = ServiceIterator(mm_eca, mm_rca)
    mm_all.set_iterator(service_iterator)

    # ping matomo at start of calculation
    from isitfit.utils import ping_matomo
    inject_timer_start = lambda context_pre: context_pre if ping_matomo("/cost/analyze/account/start") else context_pre
    mm_all.add_listener('pre', inject_timer_start)

    service_calculator_get = ServiceCalculatorGet()
    mm_all.add_listener('ec2', service_calculator_get.per_service)

    service_calculator_save = ServiceCalculatorSave()
    mm_all.add_listener('ec2', service_calculator_save.per_service)

    service_calculator_binned = ServiceCalculatorBinned()
    mm_all.add_listener('ec2', service_calculator_binned.per_service)
    mm_all.add_listener('all', service_calculator_binned.after_all)

    # update dict and return it
    # https://stackoverflow.com/a/1453013/4126114
    # inject_analyzer = lambda context_all: dict({'analyzer': service_calculator_save}, **context_all)
    # inject_analyzer = lambda context_all: dict({'calculator_binned': service_calculator_binned}, **context_all)
    # mm_all.add_listener('all', inject_analyzer)
    # 
    # service_reporter = ServiceReporterTotals()
    # service_reporter.emailTo = share_email
    # mm_all.add_listener('all', service_reporter.postprocess)
    # mm_all.add_listener('all', service_reporter.display)
    # mm_all.add_listener('all', service_reporter.email)

    # ping matomo at end of calculation
    inject_timer_end = lambda context_all: context_all if ping_matomo("/cost/analyze/account/end") else context_all
    mm_all.add_listener('all', inject_timer_end)

    # display and email
    service_reporter = ServiceReporterBinned()
    service_reporter.emailTo = share_email
    mm_all.add_listener('all', service_reporter.display)
    mm_all.add_listener('all', service_reporter.email)

    # done
    return mm_all







import boto3
import pandas as pd
import datetime as dt
import pytz

from isitfit.utils import logger

from ..utils import SECONDS_IN_ONE_DAY, myreturn, NoCloudtrailException, IsitfitCliRunnerBreakIterator

MINUTES_IN_ONE_DAY = 60*24 # 1440


from isitfit.cost.cacheManager import RedisPandas as RedisPandasCacheManager
class EventBus:
    """
    Event bus design pattern
    https://dzone.com/articles/design-patterns-event-bus
    3 event types: pre, ec2, all
    pre: event that is triggered before iterating on the resources
    ec2: event that is triggered for each resource
    all: event that is triggered after iterating on the resources

    The pre event is triggered for all listeners before iterating on the resources.
    The all event is triggered for all listeners after iterating on the resources
    The ec2 event is triggered for all listeners for each resource

    In the dzone.com jargon, listener = subscribable

    This loose coupling allows for parallelization at a later stage
    """
    def __init__(self, description, ctx):
        # description to keep track of each pipeline runner
        self.description = description
        self.set_ndays(90) # default is 90 days

        # listeners post ec2 data fetch and post all activities
        self.listeners = {'pre':[], 'ec2': [], 'all': []}

        # click context for errors
        self.ctx = ctx


    def set_ndays(self, ndays):
        self.ndays = ndays

        # set start/end dates
        dt_now_d=dt.datetime.now().replace(tzinfo=pytz.utc)
        self.StartTime=dt_now_d - dt.timedelta(days=self.ndays)
        self.EndTime=dt_now_d
        logger.debug("Metrics start..end: %s .. %s"%(self.StartTime, self.EndTime))


    def set_iterator(self, ec2_it):
        # generic iterator (iterates over regions and items)
        self.ec2_it = ec2_it


    def add_listener(self, event, listener):
      if event not in self.listeners:
        from isitfit.cli.click_descendents import IsitfitCliError
        err_msg = "Internal dev error: Event %s is not supported for listeners. Use: %s"%(event, ",".join(self.listeners.keys()))
        raise IsitfitCliError(err_msg, self.ctx)

      self.listeners[event].append(listener)


    def get_ifi(self, tqdml2_obj):
      raise Exception("Define in derived class")



class MainManager(EventBus):

    def get_ifi(self, tqdml2_obj):
        # display name of runner
        logger.info(self.description)

        # 0th pass to count
        n_ec2_total = self.ec2_it.count()

        if n_ec2_total==0:
          import click
          click.secho("No resources found in %s"%self.ec2_it.service_description, fg="red")
          return

        # context for pre listeners
        context_pre = {}
        context_pre['ec2_instances'] = self.ec2_it
        context_pre['region_include'] = self.ec2_it.get_regionInclude()
        context_pre['n_ec2_total'] = n_ec2_total
        context_pre['click_ctx'] = self.ctx
        context_pre['mainManager'] = self

        # call listeners
        for l in self.listeners['pre']:
          context_pre = l(context_pre)
          if context_pre is None:
            raise Exception("Breaking the chain is not allowed in listener/pre")

        # iterate over all ec2 instances
        sum_capacity = 0
        sum_used = 0
        df_all = []
        ec2_noCloudwatch = [] # FIXME DEPRECATED
        ec2_noCloudtrail = []

        # add some spaces for aligning the progress bars
        desc="Pass 2/2 through %s"%self.ec2_it.service_description
        desc = "%-50s"%desc

        # Edit 2019-11-12 use "initial=0" instead of "=1". Check more details in a similar note in "cloudtrail_ec2type.py"
        iter_wrap = tqdml2_obj(self.ec2_it, total=n_ec2_total, desc=desc, initial=0)
        for ec2_dict, ec2_id, ec2_launchtime, ec2_obj in iter_wrap:

          # context dict to be passed between listeners
          context_ec2 = {}
          context_ec2['mainManager'] = self
          if 'df_cat' in context_pre: context_ec2['df_cat'] = context_pre['df_cat'] # copy object between contexts
          context_ec2['ec2_dict'] = ec2_dict
          context_ec2['ec2_id'] = ec2_id
          context_ec2['ec2_launchtime'] = ec2_launchtime
          context_ec2['ec2_obj'] = ec2_obj

          try:
            # call listeners
            # Listener can return None to break out of loop,
            # i.e. to stop processing with other listeners
            for l in self.listeners['ec2']:
              context_ec2 = l(context_ec2)

              # skip rest of listeners if one of them returned None
              if context_ec2 is None:
                logger.debug("Listener %s is breaking per_resource for resource %s"%(l, ec2_id))
                break

          except NoCloudtrailException:
            ec2_noCloudtrail.append(ec2_id)

          except IsitfitCliRunnerBreakIterator as e:
            # check request for breaking from the iterator loop
            # eg for isitfit cost optimize --n=1
            logger.debug("Breaking from the per-resource iterator")
            break


        # call listeners
        #logger.info("... done")
        #logger.info("")
        #logger.info("")

        # set up context
        context_all = {}
        context_all['n_ec2_total'] = n_ec2_total
        context_all['mainManager'] = self
        context_all['region_include'] = self.ec2_it.region_include
        if 'df_cat' in context_pre: context_all['df_cat'] = context_pre['df_cat'] # copy object between contexts

        # more
        context_all['ec2_noCloudwatch'] = ec2_noCloudwatch # FIXME DEPRECATED
        context_all['ec2_noCloudtrail'] = ec2_noCloudtrail
        context_all['click_ctx'] = self.ctx

        # call listeners
        for l in self.listeners['all']:
          context_all = l(context_all)
          if context_all is None:
            raise Exception("Breaking the chain is not allowed in listener/all: %s"%str(l))

        # done
        #logger.info("")
        return context_all



class RunnerAccount(EventBus):
  def get_ifi(self, tqdml2_obj):
    if len(self.listeners['pre']) > 0:
        context_pre = {}
        # call listeners
        for l in self.listeners['pre']:
          context_pre = l(context_pre)
          if context_pre is None:
            raise Exception("Breaking the chain is not allowed in listener/pre")

    if len(self.listeners['ec2']) > 0:
        # iterate over services
        n_service_total = self.ec2_it.count()
        iter_wrap = tqdml2_obj(self.ec2_it, total=n_service_total, desc=self.ec2_it.service_description, initial=0)
        for ec2_dict, ec2_id, ec2_launchtime, ec2_obj in iter_wrap:

          # context dict to be passed between listeners
          context_ec2 = {}
          context_ec2['ec2_id'] = ec2_id
          context_ec2['ec2_obj'] = ec2_obj

          # call listeners
          # Listener can return None to break out of loop,
          # i.e. to stop processing with other listeners
          for l in self.listeners['ec2']:
            context_ec2 = l(context_ec2)

            # skip rest of listeners if one of them returned None
            if context_ec2 is None: break

    if len(self.listeners['all']) > 0:
        # set up context
        context_all = {}
        context_all['click_ctx'] = self.ctx

        # call listeners
        for l in self.listeners['all']:
          context_all = l(context_all)
          if context_all is None:
            raise Exception("Breaking the chain is not allowed in listener/all: %s"%str(l))




from .apiMan import ApiMan
import json
from isitfit.utils import logger


from isitfit.cli.click_descendents import IsitfitCliError




class EmailMan:

  def __init__(self, dataType, dataVal, ctx):
    self.dataType = dataType
    self.dataVal = dataVal
    self.ctx = ctx
    self.api_man = ApiMan(tryAgainIn=1, ctx=ctx)
    self.try_again = 3 # max attempts to try again

    # for set/get last used email
    from isitfit.dotMan import DotLastEmail, DotMan
    self.last_email = DotLastEmail()

    # get uuid
    self.isitfit_uuid = DotMan().get_myuid()


  def _send_core(self, share_email):
    # submit POST http request
    response_json, dt_now = self.api_man.request(
      method='post',
      relative_url='./share/email',
      payload_json={
        'isitfit_uuid': self.isitfit_uuid,
        'dataType': self.dataType,
        'dataVal': self.dataVal,
        'share_email': share_email
      },
      # Update 2019-12-13 change this endpoint to become anonymous
      #authenticated_user_path=True,
      #anonymous_user_path=False
      authenticated_user_path=False,
      anonymous_user_path=True
    )
    return response_json, dt_now


  def send(self, share_email):
    if share_email is None:
      return

    if len(share_email)==0:
      return

    logger.info("Sending email")

    # pre-process spaces if any
    share_email = [x.strip() for x in share_email]

    # get resources available
    # Update 2019-12-13 no need to register with API after istifit-api /share/email became anonymous
    # self.api_man.register()

    # submit POST http request
    response_json, dt_now = self._send_core(share_email)

    # check response status

    # Update 2019-12-12 Instead of raising an exception and aborting,
    # show the user a prompt to check his/her email
    # and give the program a chance to re-send the email
    import click
    while (
        (response_json['isitfitapi_status']['code']=='Email verification in progress')
        and (self.try_again > 0)
      ):
        # https://click.palletsprojects.com/en/7.x/api/?highlight=click%20confirm#click.pause
        click.pause(info='A verification link was emailed to you now. Please click the link, then press any key here to continue...', err=False)
        self.try_again -= 1
        response_json, dt_now = self._send_core(share_email)

    if self.try_again==0:
        raise IsitfitCliError(response_json['isitfitapi_status']['description'], self.ctx)

    # Update 2019-12-12 This code will get handled by apiMan and will never arrive here, so commenting it out
    #if response_json['isitfitapi_status']['code']=='error':
    #    raise IsitfitCliError(response_json['isitfitapi_status']['description'], self.ctx)

    if response_json['isitfitapi_status']['code']!='ok':
        response_str = json.dumps(response_json)
        raise IsitfitCliError("Unsupported response from server: %s"%response_str, self.ctx)

    # Save the 1st entry as "last-used email"
    # Make sure to save this *after* the verification steps above are done so as to maintain the last *working* email
    self.last_email.set(share_email[0])

    # validate schema
    from schema import SchemaError, Schema, Optional
    register_schema_2 = Schema({
        'from': str,
        Optional(str): object
      })
    try:
        register_schema_2.validate(response_json['isitfitapi_body'])
    except SchemaError as e:
        responseBody_str = json.dumps(response_json['isitfitapi_body'])
        err_msg = "Received response body: %s. Schema error: %s"%(responseBody_str, str(e))
        raise IsitfitCliError(err_msg, self.ctx)

    # otherwise proceed
    emailFrom = response_json['isitfitapi_body']['from']
    import click
    click.echo("Email sent from %s to: %s"%(emailFrom, ", ".join(share_email)))
    return




import pytest

# https://stackoverflow.com/a/40884671/4126114
@pytest.mark.parametrize("real_dsn", ["https://foo@webhook.site/123456", 'https://api-dev.isitfit.io/v0/fwd/sentry'])
def test_sentryProxy(real_dsn):
    # init now with the object
    print("Test")
    from isitfit.sentry_proxy import init
    init(real_dsn)

    # trigger
    from sentry_sdk import capture_exception
    try:
        #1/0
        bla
    except Exception as e:
        capture_exception(e)

    assert True # no exception
    # sentry_sdk/transport.py#L135 should use the real_dsn above
    # MyDsn.to_auth should get called
    # How to test?




from isitfit.utils import logger


import click
from isitfit.cli.click_descendents import IsitfitCliError
import pandas as pd


class MigMan:
  """
  Class that manages a local sqlite database to keep track of migrations that were already run
  https://www.pythoncentral.io/introduction-to-sqlite-in-python/
  """
  def __init__(self):
    self.quiet = False
    self.not_dry_run = False

    from isitfit.dotMan import DotMan
    import os
    self.db_p = os.path.join(DotMan().get_dotisitfit(), "migrations.sqlite")
    import datetime as dt
    self.dt_now = dt.datetime.now().date()
    

  def connect(self):
    import sqlite3
    self.db_h = sqlite3.connect(self.db_p)

  def _current(self):
    # Append new migrations here after implementing them as a function with a docstring in miglist.py
    from isitfit.migrations import miglist
    return [
      ('mig20191203a', miglist.mig20191203a),
      ('mig20191203b', miglist.mig20191203b),
      ('mig20191203c', miglist.mig20191203c),
    ]

  def __exit__(self):
    self.db_h.close()

  def _create(self):
    # Note: using migname instead of name because "name" is used in pandas as the "name of the row"
    cursor = self.db_h.cursor()
    cursor.execute('''
    CREATE TABLE IF NOT EXISTS migrations(
      migname TEXT PRIMARY KEY,
      executed DATE
    )
    ''')
    self.db_h.commit()

  def read(self):
    # in case of first run
    self._create()

    # insert "new" migrations
    df_mer = self._insertNew()

    # append docstrings
    df_mer['description'] = df_mer.func.apply(lambda x: x.__doc__.strip() if x.__doc__ is not None else None)

    logger.debug("Migrations")
    logger.debug(df_mer[['migname', 'executed', 'description']])

    # subset for those that don't have an executed date yet
    df_mer = df_mer[df_mer.executed.isna()]

    # save
    self.df_mig = df_mer



  def _insertNew(self):
    # migrations in local database
    df_db = pd.read_sql_query("select * from migrations", self.db_h)

    # full list of migrations
    df_py = pd.DataFrame(self._current(), columns=['migname', 'func'])

    # join. Note the left join will drop from the database any migrations that are deleted from self._current
    df_mer = df_py.merge(df_db, on='migname', how='left')

    # save to db
    df_mer[['migname', 'executed']].to_sql('migrations', self.db_h, if_exists='replace')

    # done
    return df_mer


  def migrate_all(self):
    if self.df_mig.shape[0]==0:
      if self.quiet: return
      raise IsitfitCliError("No migrations to execute")

    for i in range(self.df_mig.shape[0]):
      self._migrate_single(i)


  def _migrate_single(self, i):
    mig_i = self.df_mig.iloc[i]

    prefix = "" if self.not_dry_run else "[Dry run] "

    if not self.quiet:
      click.echo("%sExecuting migration: %s: %s"%(prefix, mig_i.migname, mig_i.description))

    if self.not_dry_run:
      mig_i.func()
      # no need to update local dataframe
      # self.df_mig.iloc[i, self.df_mig.columns == 'executed'] = self.dt_now

      # update row in database (per migration in case of error halfway)
      cursor = self.db_h.cursor()
      cursor.execute('''
      UPDATE migrations
      SET executed = ?
      where migname = ?
      ''', (self.dt_now, mig_i.migname, ))
      self.db_h.commit()



#-----------------------------
# utility functions
#def prompt_migrate():
#  migman = MigMan()
#  migman.connect()
#  migman.read()
#  if migman.df_mig.shape[0]==0:
#    return
#
#  raise IsitfitCliError("There are %i migrations that need to be executed. Please use `isitfit migrations show` to list them"%migman.df_mig.shape[0])



def silent_migrate():
  migman = MigMan()
  migman.quiet = True
  migman.not_dry_run = True
  migman.connect()
  migman.read()
  migman.migrate_all() 
  return migman.df_mig.migname.tolist()



# redshift pricing as of 2019-11-12 in USD per hour, on-demand, ohio
# https://aws.amazon.com/redshift/pricing/
redshiftPricing_dict = {
  'dc2.large': 0.25,
  'dc2.8xlarge': 4.80,
  'ds2.xlarge': 0.85,
  'ds2.8xlarge': 6.80,
  'dc1.large': 0.25,
  'dc1.8xlarge': 4.80,
}


# convert above dict to pandas dataframe
import pandas as pd
redshiftPricing_df = [{'NodeType': k, 'Cost': v} for k, v in redshiftPricing_dict.items()]
redshiftPricing_df = pd.DataFrame(redshiftPricing_df)
#print("redshift pricing")
#print(redshiftPricing_df)



def getnow():
    # date now
    import datetime as dt
    import pytz
    dtnow = dt.datetime.utcnow().replace(tzinfo=pytz.utc)
    return dtnow


class TestSqliteManUpdateDtCreated:
  def test_ok(self, mocker):
    # date now
    import datetime as dt
    import pytz
    dtnow = dt.datetime.utcnow().replace(tzinfo=pytz.utc)

    # pandas fixture table
    import pandas as pd
    table_fixture = pd.DataFrame([
      {'service': 'EC2',
       'region': 'us-west-2',
       'resource_id': 'i-1',
       'resource_size1': 't3.medium',
       'resource_size2': None,
       'classification_1': 'Underused',
       'classification_2': 'No ram',
       'cost_3m': 106,
       'recommended_size1': 't3.small',
       'savings': -53,
       'tags': [1,2,3],
       #'dt_detected': dtnow,
      },
      {'service': 'EC2',
       'region': 'us-west-2',
       'resource_id': 'i-2',
       'resource_size1': 't3.medium',
       'resource_size2': None,
       'classification_1': 'Underused',
       'classification_2': 'No ram',
       'cost_3m': 106,
       'recommended_size1': 't3.small',
       'savings': -53,
       'tags': [1,2,3],
       #'dt_detected': dtnow,
      },
      {'service': 'EC2',
       'region': 'us-west-2',
       'resource_id': 'i-3',
       'resource_size1': 't3.medium',
       'resource_size2': None,
       'classification_1': 'Underused',
       'classification_2': 'No ram',
       'cost_3m': 106,
       'recommended_size1': 't3.small',
       'savings': -53,
       'tags': [1,2,3],
       #'dt_detected': dtnow,
      }
    ])

    # context_all
    class ClickCtx:
      obj = {
        'aws_profile': 'test-profile'
      }

    context_all = {
    }

    # work within a new tempdir
    import tempfile
    with tempfile.TemporaryDirectory(prefix="isitfit-test-") as td:
      def mockreturn(*args, **kwargs): return td
      mocked_post = mocker.patch('isitfit.dotMan.DotMan.get_dotisitfit', side_effect=mockreturn)

      # init
      import time
      from isitfit.cost.account_cost_optimize import SqliteMan
      click_ctx = ClickCtx()
      sr = SqliteMan(click_ctx)

      # 1st run, no sqlite exists yet, assert only 1 value of dt_detected
      table_fixture['dt_detected'] = getnow()
      context_all['table_c'] = table_fixture.iloc[:1].copy()
      context_all = sr.update_dtCreated(context_all)

      tc_bkp1 = context_all['table_c'].copy()
      assert tc_bkp1.shape[0] == 1
      assert len(set(tc_bkp1.dt_detected.to_list())) == 1

      # sleep x seconds
      time.sleep(2)

      # Add a new recommendation, assert 2 dates available
      table_fixture['dt_detected'] = getnow()
      context_all['table_c'] = table_fixture.iloc[:2].copy()
      context_all = sr.update_dtCreated(context_all)

      tc_bkp2 = context_all['table_c'].copy()
      assert tc_bkp2.shape[0] == 2
      assert len(set(tc_bkp2.dt_detected.to_list())) == 2

      # sleep x seconds
      time.sleep(2)

      # drop 1st 2 recommendations and create a 3rd new one, assert 1 dates available
      table_fixture['dt_detected'] = getnow()
      context_all['table_c'] = table_fixture.iloc[2:3].copy()
      context_all = sr.update_dtCreated(context_all)

      tc_bkp3 = context_all['table_c'].copy()
      assert tc_bkp3.shape[0] == 1
      assert len(set(tc_bkp3.dt_detected.to_list())) == 1

      # sleep x seconds
      time.sleep(2)

      # drop all recommendations and check code doesn't fail, assert 0 dates available
      table_fixture['dt_detected'] = getnow()
      context_all['table_c'] = table_fixture.iloc[0:0].copy()
      context_all = sr.update_dtCreated(context_all)

      tc_bkp4 = context_all['table_c'].copy()
      assert tc_bkp4.shape[0] == 0
      # assert len(set(tc_bkp4.dt_detected.to_list())) == 0 # How to deal with this?

      # sleep x seconds
      time.sleep(2)

      # bring back all 3 recomendations, assert 3 dates available (i.e. the drop-all didn't wipe out the recommendations_previous table)
      # Update: actually, sticking with the simple method to update the recommendations_previous table,
      # which in fact wipes out the previous dates, and hence shows 1 single "new" date here
      table_fixture['dt_detected'] = getnow()
      context_all['table_c'] = table_fixture.iloc[0:3].copy()
      context_all = sr.update_dtCreated(context_all)

      tc_bkp5 = context_all['table_c'].copy()
      assert tc_bkp5.shape[0] == 3
      assert len(set(tc_bkp5.dt_detected.to_list())) == 1 # not 3



from isitfit.cost.metrics_automatic import MetricsAuto

class TestMetricsAuto:
  def test_displayStatus(self):
    metrics = MetricsAuto(None, None)
    metrics.status = {
      'i-1': {'ID': 'i-1', 'datadog': 'ok', 'cloudwatch': 'ok'},
      'i-2': {'ID': 'i-2', 'datadog': 'ok', 'cloudwatch': 'ok'},
    }
    metrics.display_status()


"""
  # Tests moved from test_metricsDatadog after its per_ec2 (listener class) was deprecated in favor of the metrics_automatic listener
  # Need to uncomment these tests some day and fix them for running on the metrics_automatic listener

  def test_perEc2_notConf(self, datadog_manager, ddgenv_missing):
      ddm = datadog_manager()
      assert not ddm.is_configured()
      context_ec2 = {}
      context_ec2 = ddm.per_ec2(context_ec2)
      assert context_ec2['ddg_df'] is None


  def test_perEc2_noData(self, datadog_manager, ddgenv_set):
      class MockInst:
        instance_id = 'i-123456'

      ddm = datadog_manager()
      ddm.get_metrics_all = lambda *args, **kwargs: None

      mock_obj = MockInst()
      context_ec2 = {'ec2_obj': mock_obj}
      context_ec2 = ddm.per_ec2(context_ec2)
      assert context_ec2['ddg_df'] is None


  def test_perEc2_yesData(self, datadog_manager, ddgenv_set):
      class MockInst:
        instance_id = 'i-123456'

      import datetime as dt
      ddg_met = pd.DataFrame({'ts_dt': [dt.datetime.now()]})
      ddm = datadog_manager()
      ddm.get_metrics_all = lambda *args, **kwargs: ddg_met

      mock_obj = MockInst()
      ec2_df = pd.DataFrame({'Timestamp': []})
      context_ec2 = {'ec2_obj': mock_obj, 'ec2_df': ec2_df}
      context_ec2 = ddm.per_ec2(context_ec2)
      assert context_ec2['ddg_df'].shape[0]==1
      assert context_ec2['ddg_df'].shape[1]==1
"""



"""
Manager class for getting datadog data for isitfit

Pre-requisites
    pip3 install datadog asciiplotlib pandas
    apt-get install gnuplot

set env vars
    export DATADOG_API_KEY=...
    export DATADOG_APP_KEY=...
    
Run tests
    pip3 install pytest
    pytest datadogManager.py
"""

import json
import time
import pandas as pd
import os

from isitfit.utils import logger


# Configure the module according to your needs
# Use Datadog REST API client
import datadog

from isitfit.utils import SECONDS_IN_ONE_DAY, DdgNoData, HostNotFoundInDdg, DataNotFoundForHostInDdg


class DataQueryError(ValueError):
  pass


class DatadogApiWrap:
    """
    Assisting the datadog assistant
    """

    def map_aws_datadog(self):
      # Build a map from AWS ID to Datadog hostname
      # FIXME Should probably paginate in pages of 100 hosts using combinations of start and count
      # Leaving for later until this proves to be a problem for isitfit memory consumption during execution.
      # https://docs.datadoghq.com/api/?lang=python#search-hosts
      # https://docs.datadoghq.com/agent/faq/how-datadog-agent-determines-the-hostname/?tab=agentv6v7#potential-host-names
      MAX_COUNT = 1000 # safety net, will cause problems for larger infra
      h_rev = datadog.api.Hosts.search(count=MAX_COUNT)
      # alternatively, can use host_name here.
      # Note the similar field used in self.hosts_search below.
      # If this field is changed from name to host_name, remember to change it below also
      h_rev = {x['aws_id']: x['name'] for x in h_rev['host_list'] if 'aws_id' in x and 'name' in x}
      return h_rev


    def hosts_search(self, dd_hostname):
      # https://docs.datadoghq.com/api/?lang=bash#search-hosts
      h_all = datadog.api.Hosts.search(filter='host:%s'%dd_hostname)

      # check if found
      if len(h_all['host_list'])==0:
          raise HostNotFoundInDdg("Did not find datadog hostname %s"%dd_hostname)

      # check if found
      # deprecate above? (of course change condition to != 1)
      if h_all['total_returned']>1:
          raise HostNotFoundInDdg("Found multiple hosts identified by datadog hostname %s"%dd_hostname)

      # filter for the dd_hostname
      # Not enough to take the first because I had a bug where 
      # h_all = api.Hosts.search(host=self.dd_hostname)
      # was returning a list of all hosts because the argument host=... is not supported
      # I'm not sure where I got this usage pattern from, maybe deprecated code.
      # To make matters worse, the datadog api allows the user to pass any parameters 
      # without any checks on their validity: datadog/api/resources.py SearchableAPIResource._search just passes **params to the API
      # WRONG # h_i = h_all['host_list'][0]
      # Alternatively can use host_name, not sure if one is preferred or if one will prove to be buggy
      # Will see how things come out of issue #10
      # https://github.com/autofitcloud/isitfit/issues/10
      # where name was a FQDN instead of an AWS ID
      # For more info, check https://docs.datadoghq.com/agent/faq/how-datadog-agent-determines-the-hostname/?tab=agentv6v7#potential-host-names
      # Note the similar field used in self.map_aws_datadog above.
      # If this field is changed from name to host_name, remember to change it above also
      h_i = [x for x in h_all['host_list'] if x['name']==dd_hostname]

      if len(h_i)==0:
        raise HostNotFoundInDdg("Datadog API usage problem. Hosts.search returns set of hosts, but filtering on instance ID returns an empty set. Applied to host %s in datadog"%dd_hostname)

      # at this stage we can take the first entry
      h_i = h_i[0]

      # 
      gohai = json.loads(h_i['meta']['gohai'])
      memory_total = int(gohai['memory']['total'].replace('kB',''))*1024
      out = {'cpuCores': h_i['meta']['cpuCores'], 'memory_total': memory_total}
      return out


    def metric_query(self, dd_hostname, start, end, query, metric_name, dfcol_name):
      """
      start - number of seconds since unix epoch
      end   - similar
      """
      # convert to seconds since unix epoch
      # https://stackoverflow.com/a/6999787/4126114
      # if start/end were like dt.datetime.now(), then this would be needed
      # but since they're integers to begin with (check DatadogManager.set_ndays, no need for these
      #conv2sec = lambda x: time.mktime(x.timetuple())
      #ue_start = conv2sec(start)
      #ue_end   = conv2sec(end)
      ue_start = start
      ue_end = end

      # query datadog
      # https://docs.datadoghq.com/api/?lang=python#query-timeseries-points
      m = datadog.api.Metric.query(start=ue_start, end=ue_end, query=query)

      if m['status'] != 'ok':
        raise DataQueryError(m['status'])

      if 'errors' in m:
        raise DataQueryError(m['errors'])

      if m['status']!='ok':
        msg1 = m['status']
        msg2 = 'n/a'
        if 'error' in m: msg2 = m['error']
        msg3 = "%s: %s"%(msg1, msg2)
        raise DataQueryError(msg3)

      if len(m['series'])==0:
          raise DataNotFoundForHostInDdg("No %s found for %s"%(metric_name, dd_hostname))

      # filter for the series
      # Instead of just doing
      #    m = m['series'][0]
      # which is not safe (similar to the issue I had with hosts above being a list of other IDs)
      # filter on the metric name
      m = [x for x in m['series'] if x['metric']==metric_name]

      if len(m)==0:
        raise DataNotFoundForHostInDdg("Datadog api usage. Metric query returns set of metrics not corresponding to request. Found for metric=%s and host=%s"%(metric_name, dd_hostname))

      # now safer to filter
      m = m[0]

      # convert to pandas dataframe
      df = pd.DataFrame(m['pointlist'], columns=['ts_int', dfcol_name])
      df['ts_dt'] = pd.to_datetime(df.ts_int, origin='unix', unit='ms')
      del df['ts_int']
      return df


class DatadogAssistant:
    def __init__(self, start, end, dd_hostname):
        self.end = end
        self.start = start
        self.dd_hostname = dd_hostname
        self.apiwrap = DatadogApiWrap()


    def _get_metrics_core(self, query, metric_name, col_i):
        return self.apiwrap.metric_query(dd_hostname=self.dd_hostname, start=self.start, end=self.end, query=query, metric_name=metric_name, dfcol_name=col_i)

    def _get_meta(self):
        return self.apiwrap.hosts_search(self.dd_hostname)
        
    def get_metrics_cpu_max(self):
        # query language
        # https://docs.datadoghq.com/graphing/functions/
        # Use minimum so that cpu_used will be the maximum
        query = 'system.cpu.idle{host:%s}.rollup(min,%i)'%(self.dd_hostname, SECONDS_IN_ONE_DAY)
        col_i = 'cpu_idle_min'
        metric_name = 'system.cpu.idle'
        df = self._get_metrics_core(query, metric_name, col_i)
        # calculate cpu used as 100 - cpu_idle
        df['cpu_used_max'] = 100 - df.cpu_idle_min
        df['cpu_used_max'] = df['cpu_used_max'].astype(int)
        return df

    def get_metrics_cpu_min(self):
        # query language
        # https://docs.datadoghq.com/graphing/functions/
        # Use minimum so that cpu_used will be the maximum
        query = 'system.cpu.idle{host:%s}.rollup(max,%i)'%(self.dd_hostname, SECONDS_IN_ONE_DAY)
        col_i = 'cpu_idle_max'
        metric_name = 'system.cpu.idle'
        df = self._get_metrics_core(query, metric_name, col_i)
        # calculate cpu used as 100 - cpu_idle
        df['cpu_used_min'] = 100 - df.cpu_idle_max
        df['cpu_used_min'] = df['cpu_used_min'].astype(int)
        return df
        
    def get_metrics_cpu_avg(self):
        # repeat for average
        query = 'system.cpu.idle{host:%s}.rollup(avg,%i)'%(self.dd_hostname, SECONDS_IN_ONE_DAY)
        col_i = 'cpu_idle_avg'
        metric_name = 'system.cpu.idle'
        df = self._get_metrics_core(query, metric_name, col_i)
        df['cpu_used_avg'] = 100 - df.cpu_idle_avg
        df['cpu_used_avg'] = df['cpu_used_avg'].astype(int)
        return df

    def get_metrics_ram_max(self):
        # query language, check note above in get_metrics_cpu
        query = 'system.mem.free{host:%s}.rollup(min,%i)'%(self.dd_hostname, SECONDS_IN_ONE_DAY)
        col_i = 'ram_free_min'
        metric_name = 'system.mem.free'
        df =  self._get_metrics_core(query, metric_name, col_i)
        memory_total = self._get_meta()['memory_total']
        df['ram_free_min'] = df.ram_free_min / memory_total * 100
        df['ram_free_min'] = df['ram_free_min'].astype(int)
        df['ram_used_max'] = 100 - df['ram_free_min']
        return df

    def get_metrics_ram_min(self):
        # query language, check note above in get_metrics_cpu
        query = 'system.mem.free{host:%s}.rollup(max,%i)'%(self.dd_hostname, SECONDS_IN_ONE_DAY)
        col_i = 'ram_free_max'
        metric_name = 'system.mem.free'
        df =  self._get_metrics_core(query, metric_name, col_i)
        memory_total = self._get_meta()['memory_total']
        df['ram_free_max'] = df.ram_free_max / memory_total * 100
        df['ram_free_max'] = df['ram_free_max'].astype(int)
        df['ram_used_min'] = 100 - df['ram_free_max']
        return df

    def get_metrics_ram_avg(self):
        # query language, check note above in get_metrics_cpu
        query = 'system.mem.free{host:%s}.rollup(avg,%i)'%(self.dd_hostname, SECONDS_IN_ONE_DAY)
        col_i = 'ram_free_avg'
        metric_name = 'system.mem.free'
        df =  self._get_metrics_core(query, metric_name, col_i)
        memory_total = self._get_meta()['memory_total']
        df['ram_free_avg'] = df.ram_free_avg / memory_total * 100
        df['ram_free_avg'] = df['ram_free_avg'].astype(int)
        df['ram_used_avg'] = 100 - df['ram_free_avg']
        return df

    def get_metrics_count(self):
        # query language, check note above in get_metrics_cpu
        query = 'count_not_null(system.mem.free{host:%s})'%(self.dd_hostname)
        col_i = 'nhours'
        metric_name = 'count_not_null(system.mem.free)'
        df1 =  self._get_metrics_core(query, metric_name, col_i)
        # yields data per hour, so process in pandas to daily
        df2 = df1.set_index('ts_dt').resample('1D').nhours.sum().reset_index()
        return df2


class DatadogManager:
    def __init__(self):
        datadog.initialize()
        self.set_ndays(90) # default is 90 days
        self.print_configured = True
        self.map_aws_dd = None


    def set_ndays(self, ndays):
        self.ndays = ndays
        self.end = int(time.time())
        
        # datadog will automatically set the resolution based on the start-end range
        # if the ".rollup" is not used in the query
        # number of seconds in 1 hour, if no ".rollup" specified in query, this yields 20-second frequency from datadog
        # n_secs = 60*60
        # number of seconds in 90 days, if no ".rollup" specified in query, yields bi-daily from datadog
        n_secs = SECONDS_IN_ONE_DAY*self.ndays
        self.start = self.end - n_secs


    def is_configured(self):
      # check not None and not empty string
      if os.getenv('DATADOG_API_KEY', None):
        if os.getenv('DATADOG_APP_KEY', None):
          if self.print_configured:
            logger.info("Datadog env vars available")
            self.print_configured = False
          return True
          
      if self.print_configured:
        logger.info("Datadog env vars missing. Set DATADOG_API_KEY and DATADOG_APP_KEY to get memory data from Datadog.")
        self.print_configured = False

      return False


    def build_map_aws_dd(self):
        apiwrap = DatadogApiWrap()
        self.map_aws_dd = apiwrap.map_aws_datadog()


    def get_metrics_all(self, aws_id):
        # convert aws ID to datadog hostname
        if self.map_aws_dd is None:
          self.build_map_aws_dd()
          if self.map_aws_dd is None:
            raise Exception("Failed to build aws-datadog ID map")

        # fail if not found
        if aws_id not in self.map_aws_dd:
          raise HostNotFoundInDdg("Did not find host aws ID %s in datadog reverse map"%aws_id)

        dd_hostname = self.map_aws_dd[aws_id]

        # FIXME: we already have cpu from cloudwatch, so maybe just focus on ram from datadog
        logger.debug("Fetching datadog data for aws ID %s, datadog hostname %s"%(aws_id, dd_hostname))
        ddgL2 = DatadogAssistant(self.start, self.end, dd_hostname)
        df_cpu_max = ddgL2.get_metrics_cpu_max()
        df_cpu_min = ddgL2.get_metrics_cpu_min()
        df_cpu_avg = ddgL2.get_metrics_cpu_avg()
        df_ram_max = ddgL2.get_metrics_ram_max()
        df_ram_min = ddgL2.get_metrics_ram_min()
        df_ram_avg = ddgL2.get_metrics_ram_avg()
        df_count   = ddgL2.get_metrics_count()
        df_all = (
            df_cpu_max
            .merge(df_cpu_min, how='outer', on=['ts_dt'])
            .merge(df_cpu_avg, how='outer', on=['ts_dt'])
            .merge(df_ram_max, how='outer', on=['ts_dt'])
            .merge(df_ram_min, how='outer', on=['ts_dt'])
            .merge(df_ram_avg, how='outer', on=['ts_dt'])
            .merge(df_count,   how='outer', on=['ts_dt'])
        )
        df_all = df_all[['ts_dt', 'cpu_used_max', 'cpu_used_min', 'cpu_used_avg', 'ram_used_max', 'ram_used_min', 'ram_used_avg', 'nhours']]

        # convert from datetime to date to be able to merge with cloudtrail
        df_all['ts_dt'] = df_all.ts_dt.dt.date

        # rename like cloudwatch
        df_all.rename(columns={'ts_dt': 'Timestamp'}, inplace=True)

        return df_all


from .cacheManager import MetricCacheMixin


class DatadogCached(MetricCacheMixin, DatadogManager):
    def get_key(self, aws_id):
        cache_key = "datadog:cpu+ram:%s:%i"%(aws_id, self.ndays)
        return cache_key

    def get_metrics_base(self, rc_describe_entry, rc_id, rc_created):
      return self.get_metrics_all(rc_id)


#class DatadogListener(DatadogCached):
#    """
#    A listener for the Event Bus defined in mainManager.py
#    """
#    def per_ec2(self, context_ec2):
#        raise Exception("Deprecated")
#
#        if not self.is_configured():
#          context_ec2['ddg_df'] = None
#          return context_ec2
#
#        # parse out keys
#        # FIXME build/use map from AWS EC2 instance ID to Datadog hostname
#        dd_hostname = context_ec2['ec2_obj'].instance_id
#
#        # get data
#        ddg_df = self.get_metrics_all(dd_hostname)
#
#        if ddg_df is None:
#          context_ec2['ddg_df'] = None
#          return context_ec2
#
#        # add to context
#        context_ec2['ec2_df'] = ec2_df # update context
#        context_ec2['ddg_df'] = ddg_df
#
#        # return
#        return context_ec2
#



import pandas as pd
import os
from ..utils import NoCloudtrailException

from isitfit.utils import logger



from isitfit.cost.cloudtrail_iterator import dict2service


class CloudtrailCached:
    def __init__(self, EndTime, cache_man, tqdmman):
        self.EndTime = EndTime
        self.tqdmman = tqdmman
        self.cache_man = cache_man


    def init_data(self, context_pre):
        # parse out of context
        ec2_instances, region_include, n_ec2 = context_pre['ec2_instances'], context_pre['region_include'], context_pre['n_ec2_total']

        # get cloudtail ec2 type changes for all instances
        from isitfit.cost.cloudtrail_iterator import EventAggregatorPostprocessed
        eap = EventAggregatorPostprocessed(region_include, self.tqdmman, self.cache_man, self.EndTime)
        self.df_cloudtrail = eap.get(ec2_instances, n_ec2)

        # done
        return context_pre


    def single(self, context_ec2):
        ec2_dict = context_ec2['ec2_dict']

        # imply service name
        ec2_dict['ServiceName'] = dict2service(ec2_dict)
        ServiceName = ec2_dict['ServiceName']
        region_name = ec2_dict['Region']

        sub_ct = self.df_cloudtrail

        sub_ct = sub_ct.loc[region_name]
        if sub_ct.shape[0]==0:
          raise NoCloudtrailException("No cloudtrail data #4 for %s"%ec2_id)

        sub_ct = sub_ct.loc[ServiceName]
        if sub_ct.shape[0]==0:
          raise NoCloudtrailException("No cloudtrail data #3 for %s"%ec2_id)

        # continue
        # ec2_obj = context_ec2['ec2_obj']
        ec2_id = context_ec2['ec2_id']

        # pandas series of number of cpu's available on the machine over time, past 90 days
        # series_type_ts1 = self.cloudtrail_client.get_ec2_type(ec2_obj.instance_id)
        if not ec2_id in sub_ct.index:
          raise NoCloudtrailException("No cloudtrail data #1 for %s"%ec2_id)

        df_type_ts1 = sub_ct.loc[ec2_id]
        if df_type_ts1 is None:
          raise NoCloudtrailException("No cloudtrail data #2 for %s"%ec2_id)

        # set in context
        context_ec2['df_type_ts1'] = df_type_ts1

        # done
        return context_ec2


